#!/usr/bin/env python3
import argparse
from concurrent.futures import process
import sys, os
from time import sleep
import simpleflock
from common import *
import mysql.connector
import threading
from pathlib import Path, PurePath       
from removebg import RemoveBg
# from datetime import datetime
# import time
from PIL import Image 
import shutil
import multiprocessing as mp
from urllib.request import urlopen
import requests.exceptions
import signal
MYPID = os.getpid()

SQLDBG = False
MORESQLDBG = False
DRYRUN_UPDATEDB = False
CRASHDBG = True
LOOPDBG = False
SIZEDBG = False

DBG_TO_FILE = False # super useful in cron jobs
DEFAULT_LOOP_TIME = 0.25

TABLE = "photocabine_auto_jobs_removebg"
BASE_DIR="/cygdrive/c/"
STAGING_ZIP_DIR = "/tmp/removebg/"
WWW_DIR = BASE_DIR +"www/"
PHOTOS_DIR = WWW_DIR + "photos/"
# RESOURCES_DIR = WWW_DIR + "ressources/"
MAX_CONCURRENT_THREADS = 5

TEST_TIMING = False # in prod, must be False
if TEST_TIMING: 
    ID_NET_DOWN_SLEEP_TIME = 3
    DEFAULT_DELAYED_JOB_TTL = 3
    MAX_RETRIES = 5
else:
    DEFAULT_DELAYED_JOB_TTL= ONE_HOUR_IN_SECONDS * 30
    ID_NET_DOWN_SLEEP_TIME = 300 # 5 mins, customer request
    MAX_RETRIES = int(2 * ONE_MONTH_IN_SECONDS / ID_NET_DOWN_SLEEP_TIME)  # 2 months -> 17280 retries

TEST_FORCE_TIMEOUT = False
TEST_LOCAL_REMOVEDBG = False # 

#legacy order
ERR_CODE_OK = 0
ERR_CODE_BUSY = 1
ERR_CODE_RBG_TIMEOUT = 2
ERR_CODE_RBG_API_FATAL = 3
ERR_CODE_INTERNAL_EXCEPTION_FATAL = 4
#more err codes
ERR_CODE_INTERNET_DOWN = 5
ERR_CODE_RBG_TIMEOUT_FATAL = 6
ERR_CODE_INTERNET_DOWN_FATAL = 7

err_code_hr = {
     None: None,
     ERR_CODE_OK : "ok",
     ERR_CODE_BUSY : "busy",
     ERR_CODE_RBG_TIMEOUT: "rbg timeout",
     ERR_CODE_INTERNET_DOWN: "internet down",
     ERR_CODE_RBG_API_FATAL : "rbg api",
     ERR_CODE_INTERNAL_EXCEPTION_FATAL: "internal exception",
     ERR_CODE_RBG_TIMEOUT_FATAL: "rbg timeout > ttl x retries",
     ERR_CODE_INTERNET_DOWN_FATAL: "internet down > ttl x retries"
     
}
err_repeated_to_fatal = {
     ERR_CODE_RBG_TIMEOUT : ERR_CODE_RBG_TIMEOUT_FATAL, 
     ERR_CODE_INTERNET_DOWN: ERR_CODE_INTERNET_DOWN_FATAL,
}
formats_without_alpha = [ "jpg" ]

def my_die():
    os.kill(MYPID, signal.SIGKILL)

def parse_args():
    dbg(CRASHDBG, "entering")
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('--help', '-h', help='show this help message and exit', action='store_true')
    parser.add_argument('--period', '-l', help='define loop periodicity in floating seconds', type=float, default=DEFAULT_LOOP_TIME)
    parser.add_argument('--error_log', '-e', help='define full path and filename to error_log', type=str, required=False, default="rmbg_error.log")
    parser.add_argument('--log_file', '-f', help='define full path and filename to log_file', type=str, required=False, default="rmbg.log")
    parser.add_argument('--lock_file', '-k', help='lock on lockfile', default=None)
    parser.add_argument('--lock_timeout', '-t', help='lockfile timeout', default=0)
    parser.add_argument('--verbose', '-v', help='verbose', action='store_true', default=False)
    args = parser.parse_args()

    if args.help:
        parser.print_help(sys.stdout)
        sys.exit(0)

    #dbg(CRASHDBG, "exiting")
    return args

args = parse_args()

if DBG_TO_FILE:
    from functools import partial
    dbg = partial(fdbg, args.log_file) #works well with cygwin too

def consolidate_ttl(ttl):
    if ttl <=0:
        dbg(CRASHDBG, "ttl_seconds < =0: setting it to", DEFAULT_DELAYED_JOB_TTL)
        ttl = DEFAULT_DELAYED_JOB_TTL
    return ttl

def complete_fn_path(fn):        
    fn_has_no_full_path = PurePath(fn).name == fn
    dbg(CRASHDBG, "fn is", fn, "has_no_full_path", fn_has_no_full_path)
    if fn_has_no_full_path: 
        fn_full_path = PHOTOS_DIR + fn
    else: 
        fn_full_path = fn

    return fn_full_path

def move(source_fn, dest_fn):
    #3.99: move file to destination

    dest_fn_full_path = complete_fn_path(dest_fn)

    dbg(True, f"moving  {source_fn} to {dest_fn_full_path}")
    if True:
        # sleep(2) # WARNING: TEST
        os.rename(source_fn, dest_fn_full_path) 
    else : #purely for debugging the threading/process parallelism
        import shutil
        # sleep(3) # WARNING: TEST
        shutil.copy(source_fn, dest_fn_full_path) 


def compose(staging_zip_file, my_dict): #can throw exception! when an image is incomplete
    #possibly precompose
    staging_out_fn = ""
    compose_zip_dir = None
    if my_dict["out_format"] != "zip" :  #composition required
        
        compose_zip_dir = STAGING_ZIP_DIR + Path(staging_zip_file).stem + "/"
        #3. decompress
        import zipfile
        with zipfile.ZipFile(staging_zip_file, 'r') as zip_ref:
            zip_ref.extractall(compose_zip_dir)


        im_mask = Image.open(compose_zip_dir + "alpha.png")
        dbg(SIZEDBG, "im_mask ", im_mask.size)
        #3.1. compose color and alpha
        im_color  = Image.open(compose_zip_dir + "color.jpg")
        dbg(SIZEDBG, "im_color ", im_color.size)
        im_color.putalpha(im_mask)
        im_compose = im_color
        

        #3.2 compose with background if any
        dbg(CRASHDBG, "in_bg_filename", my_dict["in_bg_filename"])
        if my_dict["in_bg_filename"] != "" and my_dict["in_bg_filename"] is not None:
            if my_dict["in_bg_filename"].startswith("http://") or my_dict["in_bg_filename"].startswith("https://"):
                
                in_bg_opener = urlopen(my_dict["in_bg_filename"]) #direct URL
            else: 
                in_bg_opener = WWW_DIR + my_dict["in_bg_filename"] # usually /cygdrive/c/www/ressources/

            im_bg = Image.open(in_bg_opener)
            dbg(SIZEDBG, "im_bg ", im_bg.size)
            im_bg_resize = im_bg.resize(im_compose.size)
            dbg(SIZEDBG, "im_bg_resize ", im_bg_resize.size)
            im_compose =Image.composite(im_compose, im_bg_resize, im_mask)
            dbg(SIZEDBG, "im_compose ", im_compose.size)
        #3.8: save at the expected format
        staging_out_fn = compose_zip_dir + Path(my_dict["out_filename"]).stem + "." + my_dict["out_format"]

        dbg(CRASHDBG, "staging_out_fn", staging_out_fn)
        if (my_dict["out_format"].lower() in formats_without_alpha) and (im_compose.mode in ("RGBA", "P")):
            dbg(CRASHDBG, f"converting to RGB before saving to {my_dict['out_format']}")
            im_compose = im_compose.convert("RGB")
        im_compose.save(staging_out_fn) 
    else:
        staging_out_fn = staging_zip_file

    #4 move file to destination
    move(staging_out_fn, my_dict['out_filename'])
    if compose_zip_dir is not None:
        # 5 clean unzipped mess
        shutil.rmtree(compose_zip_dir)

    # return staging_out_fn

    


class RemoveBGJobProcessor():
    


    def timeout(self, func, _args=(), kwds = {}):
        dbg(CRASHDBG, "entering with args", _args, "kwds", kwds)
        my_dict = _args[0]
        pool = mp.Pool(processes = 1)

        #sometimes the ttl is <=0. This is because when the delayed removebg job line has been added upfront by PHP, it doesn't know in advance the DEFAULT_DELAYED_JOB_TTL. So it puts 0. 
        my_dict['ttl_seconds'] = consolidate_ttl(my_dict['ttl_seconds']) # now we have the real ttl value also for delayed RBG jobs.

        sql_update = f"UPDATE {TABLE} SET ttl_seconds = {str(my_dict['ttl_seconds'])} WHERE id = {str(my_dict['id'])}"
        self._sql_exec(sql_update, commit=True)
        timeout = my_dict['ttl_seconds']
        
        # dbg(CRASHDBG, "1")
        result = pool.apply_async(func, args=_args, kwds=kwds) 
        try:
            # dbg(CRASHDBG, "2")
            err_code = result.get(timeout=timeout)
            # dbg(CRASHDBG, "3")

        except mp.TimeoutError:
            dbg(CRASHDBG, "timeout :(")
            pool.terminate()
            dbg(True, f"timeout is over on rbg line id {my_dict['id']}")
            err_code = ERR_CODE_RBG_TIMEOUT
            
        else:
            dbg(CRASHDBG, "in time :) ")
            pool.close()
            pool.join()            

        finally:
            err_code_updated = self.update_rbg_table_line_err_code(my_dict['id'], err_code)
            #^^^indeed err_code can be modified if it gets a err_code_rbg_timeout and timeout retries have been exceeded, or an internet_down and max down delay was exceeded
            if err_code_updated == ERR_CODE_INTERNAL_EXCEPTION_FATAL:
                dbg(CRASHDBG, f"INTERNAL ERROR: {err_code_hr[err_code_updated]},  :( FIXME")

            elif err_code_updated == ERR_CODE_INTERNET_DOWN:
                dbg(CRASHDBG, f"simple internet connection (means id_photo exists). Cool down {ID_NET_DOWN_SLEEP_TIME} seconds before rescheduling this job...")
                sleep(ID_NET_DOWN_SLEEP_TIME)
                #reinit err code to None after the sleep
                self.update_rbg_table_line_err_code(my_dict['id'], err_code=err_code_updated, worker_pid="null", promote_err_code_to_fatal=False)  #
                # my_die() #debug
            elif err_code_updated == ERR_CODE_RBG_TIMEOUT: #in case of a rbg timeout, don't even look at the phocabine_photos yet
                dbg(CRASHDBG, "simple rbg timeout")
            else:
                #update photo db in the following remaining cases
                if err_code_updated in [ ERR_CODE_RBG_API_FATAL, ERR_CODE_RBG_TIMEOUT_FATAL, ERR_CODE_INTERNET_DOWN_FATAL ] : # if it failed, consider the out filename to be the in filename (generically)
                    dbg(CRASHDBG, f"FATAL ERROR: {err_code_hr[err_code_updated]}")
                    my_dict['out_filename'] = my_dict['in_filename']
                else: 
                    dbg(CRASHDBG, "success :)")
                self.update_photos_db_if_needed(my_dict) # if this "id_photo" failed, then yes, the output filename in phocabine_photos will fallback to the input filename  

            
        dbg(CRASHDBG, "exiting")


    def __init__(self, args):
        dbg(CRASHDBG, "entering")
        self.args = args
        '''
        self.mydb = mysql.connector.connect(
            host="localhost",
            user="root",
            password="",
            database="photocabine"
        )
        '''

        # create staging zip dir folder if not exists
        os.makedirs(STAGING_ZIP_DIR, exist_ok=True)

        # initial cleanup: delete every file in it 
        path = STAGING_ZIP_DIR
        for file_name in os.listdir(path):
            # construct full file path
            file = path + file_name
            if os.path.isfile(file):
                dbg(True,'Deleting file:', file)
                os.remove(file)
        # startup sanity: remove any line related to incomplete / unterminated process_removebg.py instances (ctrl+c, power break, exception...)
        # sql_update = f"UPDATE {TABLE} SET worker_pid = NULL, error_code = NULL, error_code_hr = NULL WHERE error_code IN ( {str(ERR_CODE_BUSY)}, {str(ERR_CODE_RBG_TIMEOUT)}, {str(ERR_CODE_INTERNAL_EXCEPTION_FATAL)},  NULL)"  
        sql_update = f"UPDATE {TABLE} SET worker_pid = NULL, error_code = NULL, error_code_hr = NULL WHERE error_code IS NULL OR error_code IN ({str(ERR_CODE_BUSY)}, {str(ERR_CODE_RBG_TIMEOUT)}, {str(ERR_CODE_INTERNET_DOWN)})"          
        self._sql_exec(sql_update, commit=True)


    def _sql_exec(self, sql, commit=False):
        dbg(MORESQLDBG, "entering with sql " + sql)


        '''
        this is suboptimal but whatever. we cannot keep a self.mydb connection as it can be accessed from multiple subprocesses and this is not playing nicely.
        see this error:
            Process ForkPoolWorker-1:
            Traceback (most recent call last):
            File "/usr/lib/python3.9/multiprocessing/process.py", line 315, in _bootstrap
                self.run()
            File "/usr/lib/python3.9/multiprocessing/process.py", line 108, in run
                self._target(*self._args, **self._kwargs)
            File "/usr/lib/python3.9/multiprocessing/pool.py", line 114, in worker
                task = get()
            File "/usr/lib/python3.9/multiprocessing/queues.py", line 368, in get
                return _ForkingPickler.loads(res)
            File "/usr/lib/python3.9/multiprocessing/reduction.py", line 246, in _rebuild_socket
                fd = df.detach()
            File "/usr/lib/python3.9/multiprocessing/resource_sharer.py", line 58, in detach
                return reduction.recv_handle(conn)
            File "/usr/lib/python3.9/multiprocessing/reduction.py", line 189, in recv_handle
                return recvfds(s, 1)[0]
            File "/usr/lib/python3.9/multiprocessing/reduction.py", line 164, in recvfds
                raise RuntimeError('received %d items of ancdata' %

        '''
        mydb = mysql.connector.connect(
            host="localhost",
            user="root",
            password="",
            database="photocabine")

        c = mydb.cursor(buffered=True, dictionary=True)  # burffered so that all select lines are prefetched, allowing to immediately do updates from results from a select: https://stackoverflow.com/questions/29772337/python-mysql-connector-unread-result-found-when-using-fetchone            
        c.execute(sql)
        if commit:
            mydb.commit()
        dbg(False, "exiting with cursor "+ str(c))
        return c


    def update_rbg_table_line_err_code(self, rbg_id, err_code, worker_pid=None, promote_err_code_to_fatal=True):
        dbg(CRASHDBG, "entering")
        def create_set_line(my_dict):
            s = ""
            pref =""
            i=0
            for (k,v) in my_dict.items():
                my_v = v
                if v is None:
                    my_v ="NULL"
                elif type(v) is str: 
                    my_v = f"'{my_v}'" 
                s += pref+ f"{k}={my_v}"
                if i==0:
                    pref = ", "
            return s

        do_not_touch_worker_pid = False
        d = {}
        if promote_err_code_to_fatal and (err_code in err_repeated_to_fatal.keys()):
            cursor =  self._sql_exec(f"SELECT retries, id_photo FROM {TABLE} WHERE id = " + str(rbg_id)).fetchone()
            retries, id_photo = cursor["retries"], cursor["id_photo"]
            if retries == None:
                retries = 0
            if id_photo is not None and retries < MAX_RETRIES: #
                retries += 1 
                if err_code == ERR_CODE_RBG_TIMEOUT:
                   d["worker_pid"] = None # so that on the next loop select, this line can be found. 
                #for the case of an internet down, we will call update_rbg...() to also restore worker_pid to null at end of sleep
            else:
                #promote err_code to fatal
                do_not_touch_worker_pid = True
                dbg(CRASHDBG, "promoting error!")
                err_code = err_repeated_to_fatal[err_code] # FAIL : either an id_photo that failed MAX_RETRIES times in a row OR a non-id_photo that failed once (and only once) 
            d["retries"] = retries

        if worker_pid is not None and not do_not_touch_worker_pid:
            if isinstance(worker_pid, str) and  worker_pid.lower() == "null": 
                d['worker_pid'] = None
            else:
                d["worker_pid"] = worker_pid

        d["error_code"] = err_code
        d["error_code_hr"] = err_code_hr[err_code]
        
        where_str = " WHERE id = " + str(rbg_id)
        sql_update = "UPDATE " + TABLE + " SET " + create_set_line(d) + where_str

        if not DRYRUN_UPDATEDB:
            dbg(SQLDBG, f"exiting with update query {sql_update}")
            self._sql_exec(sql_update, commit=True)
        dbg(CRASHDBG, "exiting with err_code", err_code)
        return err_code

    def update_photos_db_if_needed(self, my_dict):
        if DRYRUN_UPDATEDB or my_dict["id_photo"] is None:
            return
        dbg(CRASHDBG, "updating photocabine_photos for id_photo", my_dict["id_photo"])
        sql_update = "UPDATE phocabine_photos SET sync= NULL, " + my_dict["out_field_name"] +"=\"" + my_dict["out_filename"] + "\" WHERE id_photo = \"" + str(my_dict["id_photo"]) +"\""
        dbg(SQLDBG, sql_update)
        self._sql_exec(sql_update, commit=True)
            
    def call_removebg_and_update_db(self, *tupl): # thread method

        my_dict = tupl[0]
        dbg(CRASHDBG, "entering") # with tupl", tupl, "dict", my_dict)
        
        # inside a cron, this job could be niced too much and cause timeouts. let's unnice it to the max
        my_nice = os.nice(0)
        dbg(CRASHDBG, "initial nice: ", my_nice)
        my_nice = os.nice(-20 - my_nice)
        dbg(CRASHDBG, "new nice: ", my_nice)
        #dbg(CRASHDBG,"rbg_id", my_dict["id"], "date", my_dict["date"], "worker_pid", my_dict["worker_pid"], "api_key", my_dict["api_key"], "id_photo", my_dict["id_photo"], "in_filename", my_dict["in_filename"], "in_bg_filename", my_dict["in_bg_filename"], "out_format", my_dict["out_format"], "out_filename", my_dict["out_filename"], "out_field_name", my_dict["out_field_name"], "ttl_seconds", my_dict["ttl_seconds"])
        #dbg(CRASHDBG,"ttl seconds", my_dict["ttl_seconds"])
        
        #1 call removebg
        try: 
            in_fn_full_path = complete_fn_path(my_dict["in_filename"])
            
            base_out_zip_file = Path(my_dict["in_filename"]).stem + "_removebg.zip"
            staging_zip_file = STAGING_ZIP_DIR + base_out_zip_file
            
            if TEST_LOCAL_REMOVEDBG: 
                dbg(CRASHDBG, "TEST_LOCAL_REMOVEBG")
                #check internet connection
                if my_ping("apps.laphotocabine.com"):
                    if TEST_FORCE_TIMEOUT:
                        sleep(my_dict["ttl_seconds"] + 2)
                    shutil.copyfile(PHOTOS_DIR + "color_and_alpha.zip", staging_zip_file)
                    err_code = ERR_CODE_OK
                else:
                    err_code = ERR_CODE_INTERNET_DOWN
            else: 
                # dbg(CRASHDBG, "about to rmbg")
                if my_dict["api_key"] == "":
                    dbg(True, "NO API KEY PROVIDED FOR JOB", my_dict) # avoid calling removebg knowing that this job misses an api key
                    err_code = ERR_CODE_RBG_API_FATAL
                else:
                    rmbg = RemoveBg(my_dict["api_key"], self.args.error_log)
                    # dbg(CRASHDBG, "rmbg instance is ", rmbg)
                    rmbg.remove_background_from_img_file(in_fn_full_path, size='full', format='zip', new_file_name=staging_zip_file) 
                    # dbg(CRASHDBG, "rmbg called successfully")
                    err_code = ERR_CODE_OK

            if err_code == ERR_CODE_OK:
                #3 compose
                compose(staging_zip_file, my_dict) #this can throw when reaching an incomplete image
                
                #4 update the photocabine_base_jobs_removebg database.             
                self.valid_processed_photos +=1

                


        except requests.exceptions.ConnectionError as exc:
            dbg(True, "No internet connection:", exc)
            if my_dict["id_photo"] is not None: # let's give it a retry
                err_code = ERR_CODE_INTERNET_DOWN 
            else:
                err_code = ERR_CODE_INTERNET_DOWN_FATAL #no dice, fatal error

        except Exception as exc:             
            # dbg(True, "exception occured: ", exc)
            template = "An exception of type {0} occurred. Arguments: {1!r}"
            message = template.format(type(exc).__name__, exc.args)
            dbg(True, message)
            err_code = ERR_CODE_INTERNAL_EXCEPTION_FATAL

        finally:
            #5 delete zip file
            try:
                os.remove(staging_zip_file)
            except OSError as ex:
                pass # the file might be here, might not be, let's delete it or ask forgiveness

        return err_code
            

    def loop(self):
        dbg(CRASHDBG, "entering")
        where = f"( error_code IS NULL or error_code IN ( {str(ERR_CODE_RBG_TIMEOUT)}, {str(ERR_CODE_INTERNET_DOWN)} ) ) AND worker_pid IS NULL"
        sql_query = f"SELECT * FROM  {TABLE} WHERE {where} ORDER BY id_photo ASC LIMIT 2" #prefer when there is no id_photo == immediate job 
        while True:
            # req_start_time = get_time()
            # now = datetime.now()
            # now_str = now.strftime("%Y/%m/%d %H:%M:%S")
            
            #let's look at the lines we need to process. either virgin lines (err_code == NULL ) or lines that timed out but not yet fully failed after the retry count (err_count == ERR_CODE_RBG_TIMEOUT) 
            dbg(MORESQLDBG, sql_query)
            # we don't need the following complement in SQL query as it would prevent us to process old-dated lines... 
            # this is irrelevant but it is a good symtax hint:  "AND TIMESTAMPADD(SECOND, ttl_seconds, date) > NOW()" # see  https://www.w3resource.com/mysql/date-and-time-functions/mysql-timestampadd-function.php
            # dbg(SQLDBG or self.args.verbose, "SQL SELECT: " + sql_query)
            mycursor = self._sql_exec(sql_query)

            self.paralellized_photos_count = 0
            self.valid_processed_photos = 0
            processed_photos = 0           
            for my_dict in mycursor:
                # dbg(CRASHDBG, "my_dict is", my_dict)

                #step synchronously mark line as being in used, right now, by our pid
                self.update_rbg_table_line_err_code(my_dict['id'], err_code=ERR_CODE_BUSY, worker_pid=MYPID)
                if False:
                    self.timeout(self.call_removebg_and_update_db, args=[my_dict])
                else:
                    t = threading.Thread(target=self.timeout, args=[self.call_removebg_and_update_db],  kwargs={'_args':[my_dict]}, daemon=True)
                    t.start()
                processed_photos += 1
            self.paralellized_photos_count += processed_photos
                    
        
            # req_end_time = get_time()
            # req_duration = req_end_time - req_start_time
            dbg(LOOPDBG, f"this loop processed {processed_photos} jobs in // ; total {self.paralellized_photos_count}, total successful {self.valid_processed_photos}")
            if self.args.period is None:
                break
            
            dbg(self.args.verbose, f"sleeping {self.args.period} seconds")
            sleep(self.args.period)   #cool off the loop. threads are working in //
            

def go(args):
    rbg_proc = RemoveBGJobProcessor(args)
    rbg_proc.loop()

def main():       
    # print(args)
    if args.lock_file:
        try:
            dbg(args.verbose or CRASHDBG, f"trying to lock on {args.lock_file} with timeout {args.lock_timeout}")
            with simpleflock.SimpleFlock(args.lock_file, args.lock_timeout):
                go(args)
        except BlockingIOError as ex:
            dbg(True, f"cannot lock {args.lock_file}: already in use, good news")
            sys.exit(1)
    else: 
        go(args)


main()



